@@ -1,4 +1,4 @@
-from .jdconstants import *
+from .constants import *
 from .events import *
 from .packet import *
 from .bus import *
