@@ -3,19 +3,20 @@ import random
 import asyncio
 import queue
 import os
+import time
+import sys
 
 from typing import Optional, TypeVar, Union, cast
 
 from .constants import *
+from .control.constants import *
 from .system.constants import *
+from .unique_brain.constants import *
 from .events import *
 from .packet import *
-from .transport import Transport
 
-import jacdac.util as util
-from .util import now, log, logv, unpack
-from .control.constants import *
-from .pack import PackType, jdpack, jdunpack
+from .util import now, log, logv
+from .pack import PackTuple, PackType, jdpack, jdunpack
 
 
 EV_CHANGE = "change"
@@ -37,7 +38,7 @@ EV_DISCONNECTED = "disconnected"
 _ACK_RETRIES = const(4)
 _ACK_DELAY = const(40)
 
-RegType = TypeVar('RegType', bound=Union[int, tuple[int, ...]])
+RegType = TypeVar('RegType', bound=Union[PackType, PackTuple])
 
 
 def _service_matches(dev: 'Device', serv: bytearray):
@@ -50,9 +51,19 @@ def _service_matches(dev: 'Device', serv: bytearray):
     return True
 
 
+class Transport:
+    """A base class for packet transports"""
+
+    on_receive: Optional[Callable[[bytes], None]] = None
+    # Callback to report a received packet to the bus
+
+    def send(self, pkt: bytes) -> None:
+        # send a packet payload over the transport layer
+        pass
+
+
 class Bus(EventEmitter):
-    """A Jacdac bus that managed devices, service client, registers...
-    """
+    """A Jacdac bus that managed devices, service client, registers."""
 
     def __init__(self, transport: Transport, *, device_id: str = None) -> None:
         super().__init__(self)
@@ -100,11 +111,9 @@ class Bus(EventEmitter):
         asyncio.set_event_loop(loop)
 
         # TODO: what's the best way to import these things
-        from .control.server import ControlServer
         ctrls = ControlServer(self)  # attach control server
 
         # TODO: make this optional.
-        from .unique_brain.server import UniqueBrainServer
         brain = UniqueBrainServer(self)
 
         def keep_task(t: asyncio.Task[None]):
@@ -334,29 +343,30 @@ class RawRegisterClient(EventEmitter):
             return self._data
         return None
 
-    def unpacked(self) -> tuple[Any, ...]:
+    def values(self) -> Optional[PackTuple]:
         data = self.query_no_wait()
         if data and self.pack_format:
-            return unpack(data, self.pack_format)
-        return ()
+            return jdunpack(data, self.pack_format)
+        return None
 
-    def value(self, index: int):
-        values = self.unpacked()
-        if (len(values) > index):
-            return values[index]
-        else:
-            return None
+    def set_values(self, *args: PackType):
+        if self.pack_format is None:
+            raise RuntimeError("set_value not supported")
+        data = jdpack(self.pack_format, *args)
 
-    def set_value(self, index: int, value: Any):
-        # TODO:
-        pass
+        def send():
+            pkt = JDPacket(cmd=JD_SET(self.code), data=data)
+            self.client.send_cmd(pkt)
+            self.refresh()
 
-    def float_value(self, index: int = 0, scale: int = 1) -> Union[float, None]:
-        value = self.value(index)
-        if value is None:
+        self.bus.run(send)
+
+    def float_value(self, scale: int = 1) -> Union[float, None]:
+        values = self.values()
+        if values is None:
             return None
         else:
-            return float(value) * scale
+            return float(values[0]) * scale  # type: ignore
 
     def _query(self):
         pkt = JDPacket(cmd=JD_GET(self.code))
@@ -481,10 +491,10 @@ class Server(EventEmitter):
         self.handle_reg_u32(pkt, JD_REG_STATUS_CODE, self._status_code)
 
     def handle_reg_u32(self, pkt: JDPacket, register: int, current: int):
-        return self.handle_reg(pkt, register, "I", current)
+        return self.handle_reg(pkt, register, "u32", current)
 
     def handle_reg_i32(self, pkt: JDPacket, register: int, current: int):
-        return self.handle_reg(pkt, register, "i", current)
+        return self.handle_reg(pkt, register, "i32", current)
 
     def handle_reg(self, pkt: JDPacket, register: int, fmt: str, current: RegType) -> RegType:
         getset = pkt.service_command >> 12
@@ -494,8 +504,12 @@ class Server(EventEmitter):
         if reg != register:
             return current
         if getset == 1:
-            self.send_report(JDPacket.packed(
-                pkt.service_command, fmt, current))
+            if isinstance(current, tuple):
+                self.send_report(JDPacket.packed(
+                    pkt.service_command, fmt, *current))
+            else:
+                self.send_report(JDPacket.packed(
+                    pkt.service_command, fmt, current))
         else:
             if register >> 8 == 0x1:
                 return current  # read-only
@@ -517,7 +531,85 @@ class Server(EventEmitter):
         log(prefix + text, *args)
 
 
+class ControlServer(Server):
+    """A server for the control service, used internally by the bus."""
+
+    def __init__(self, bus: Bus) -> None:
+        super().__init__(bus, JD_SERVICE_CLASS_CONTROL)
+        self.restart_counter = 0
+
+    def queue_announce(self):
+        logv("announce: %d " % self.restart_counter)
+        self.restart_counter += 1
+        ids = [s.service_class for s in self.bus. servers]
+        rest = self.restart_counter
+        if rest > 0xf:
+            rest = 0xf
+        ids[0] = (
+            rest |
+            JD_CONTROL_ANNOUNCE_FLAGS_IS_CLIENT |
+            JD_CONTROL_ANNOUNCE_FLAGS_SUPPORTS_ACK |
+            JD_CONTROL_ANNOUNCE_FLAGS_SUPPORTS_BROADCAST |
+            JD_CONTROL_ANNOUNCE_FLAGS_SUPPORTS_FRAMES
+        )
+        buf = jdpack("u32[]", *ids)
+        self.send_report(JDPacket(cmd=0, data=buf))
+
+        # auto bind
+        # if jacdac.role_manager_server.auto_bind:
+        #     self.auto_bind_cnt++
+        #     # also, only do it every two announces (TBD)
+        #     if self.auto_bind_cnt >= 2:
+        #         self.auto_bind_cnt = 0
+        #         jacdac.role_manager_server.bind_roles()
+
+    # def handle_flood_ping(self, pkt: JDPacket):
+    #     num_responses, counter, size = pkt.unpack("IIB")
+    #     payload = bytearray(4 + size)
+    #     for i in range(size): payload[4+i]=i
+    #     def queue_ping():
+    #         if num_responses <= 0:
+    #             control.internal_on_event(
+    #                 jacdac.__physId(),
+    #                 EVT_TX_EMPTY,
+    #                 do_nothing
+    #             )
+    #         else:
+    #             payload.set_number(NumberFormat.UInt32LE, 0, counter)
+    #             self.send_report(
+    #                 JDPacket.from(ControlCmd.FloodPing, payload)
+    #             )
+    #             num_responses--
+    #             counter++
+    #     control.internal_on_event(jacdac.__physId(), EVT_TX_EMPTY, queue_ping)
+    #     queue_ping()
+
+    def handle_packet(self, pkt: JDPacket):
+        if pkt.is_reg_get:
+            if pkt.reg_code == JD_CONTROL_REG_UPTIME:
+                self.send_report(JDPacket.packed(
+                    JD_GET(JD_CONTROL_REG_UPTIME), "u64",  time.monotonic_ns() // 1000))
+        else:
+            cmd = pkt.service_command
+            if cmd == JD_CONTROL_CMD_SERVICES:
+                self.queue_announce()
+            elif cmd == JD_CONTROL_CMD_IDENTIFY:
+                self.log("identify")
+                self.bus.emit(EV_IDENTIFY)
+            elif cmd == JD_CONTROL_CMD_RESET:
+                sys.exit()  # TODO?
+
+
+class UniqueBrainServer(Server):
+    """A server for the unique brain service, used internally by the bus"""
+
+    def __init__(self, bus: Bus) -> None:
+        super().__init__(bus, JD_SERVICE_CLASS_UNIQUE_BRAIN)
+
+
 class Client(EventEmitter):
+    """Base class to define service clients."""
+
     def __init__(self, bus: Bus, service_class: int, pack_formats: dict[int, str], role: str) -> None:
         super().__init__(bus)
         self.broadcast = False
@@ -571,14 +663,14 @@ class Client(EventEmitter):
         pkt._header[3] |= JD_FRAME_FLAG_COMMAND
         self.bus._send_core(pkt)
 
-    def send_cmd_packed(self, cmd: int, args: PackType = None):
+    def send_cmd_packed(self, cmd: int, *args: PackType):
         if args is None:
             pkt = JDPacket(cmd=cmd)
         else:
             if not cmd in self.pack_formats:
                 raise RuntimeError("unknown data format")
             fmt = self.pack_formats[cmd]
-            data = jdpack(fmt, args)
+            data = jdpack(fmt, *args)
             pkt = JDPacket(cmd=cmd, data=data)
         self.send_cmd(pkt)
 
@@ -634,6 +726,8 @@ _JD_CONTROL_ANNOUNCE_FLAGS_RESTART_COUNTER_STEADY = const(0xf)
 
 
 class Device(EventEmitter):
+    """A device on the bus"""
+
     def __init__(self, bus: Bus, device_id: str, services: bytearray) -> None:
         super().__init__(bus)
         self.device_id = device_id
